/* This work is licensed under the Creative Commons Attribution-Share Alike 2.5 Canada License.
 * This To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/2.5/ca/
 * This or send a letter to Creative Commons, 171 Second Street, Suite 300, San Francisco,
 * This California, 94105, USA.
 *
 * Written by John Simpson, Department of Philosophy, University of Alberta, 2009, 2010.
*/

#include <iostream>
#include "agent.h"
#include <gsl/gsl_rng.h>
#include <algorithm>

using namespace std;

agent::agent(const int spawnDisc, const int playDisc, const bool rndLike, const bool rndAft, const int genNum, const int mA, gsl_rng * r, const int agentCount)
{
	//Increment the ID assignment variable and set the agent's ID to this new value
	id = agentCount;

	//set birth generation
	birthday=genNum;

	//set age to zero
	age = 0;

	//Set initial resources to zero
	resource = 0;

	//If an agent doesn't see a given agent for maxAbsent rounds then they are forgotten.
	//set maxAbsent to some reasonable size.  Ideally something equivalent to a generation or two.
	maxAbsent = mA;

	//Randomize an initial set of agent behaviours

	bitsSet=0;

	for (int i=0; i<(int)choice.size(); i++)
		if (gsl_rng_uniform (r) <= 0.5)
		{
			choice.set(i);
			bitsSet++;
		}

	//Randomize initial agent after play behaviour

	for (int i=0; i<3; i++)
	for (int j=0; j<3; j++)
	for (int k=0; k<3; k++)
	{
		if (rndAft==true)
		{
			if (gsl_rng_uniform (r) < 1.0/3.0)
				afterPlay[i][j][k]='E';
			else
			{
				if (gsl_rng_uniform (r) > 0.5)
					afterPlay[i][j][k]='N';
				else
					afterPlay[i][j][k]='F';
			}
		}
		else
			afterPlay[i][j][k]='N';
	}

	if (spawnDisc>0) //Randomize agent's spawning preferences
	{
		for (int i=0; i<int(spawnPref.size()); i++)
			if (gsl_rng_uniform (r) <= 0.5)
				spawnPref.set(i);
	}
	else  //Agent will spawn with anyone
	{
		for (int i=0; i<int(spawnPref.size()); i++)
			spawnPref.set(i);
	}

	if (playDisc>0) //Randomize agent's playing preferences
	{
		for (int i=0; i<int(playPref.size()); i++)
			if (gsl_rng_uniform (r) <= 0.5)
				playPref.set(i);
	}
	else  //Agent will play with anyone
	{
		for (int i=0; i<int(playPref.size()); i++)
			playPref.set(i);
	}

	if (rndLike==true)//Randomize default likings
	{
		if (gsl_rng_uniform (r) < 1.0/3.0)
				dmLu='E';
		else
			if (gsl_rng_uniform (r) > 0.5)
				dmLu='N';
			else
				dmLu='F';

		if (gsl_rng_uniform (r) < 1.0/3.0)
				duLm='E';
		else
			if (gsl_rng_uniform (r) > 0.5)
				duLm='N';
			else
				duLm='F';
	}
	else
	{
		dmLu='N';
		duLm='N';
	}

	for (int i=0; i<6; i++)
		likingStats[i]=0;

}

/* Agent's determine their "play" for a given game as follows:
 * 1. Look over the game and determine a set of decision switches based on the payoffs
 * available, the current status quo position, and what the perceived relationship is between
 * the deciding agent and the other player.
 * 2. These decision switches are then used to index a choice to either play on or off the
 * currently presented status quo.
 * 3. Play on the status quo returns "true".
*/

//!!!!! THIS SHOULD ALL BE DONE BY CLASS "GAME", AGENT THEN JUST ASKS
// Except for mLu and uLm

//Remember to pass the gamePtr as something like &gameidentifier

bool agent::play(game *const gamePtr, const int opID, const bool row)
{
	int mLu=-1;		//Knowledge agent has of other agent
	int uLm=-1;		//Knowledge agent thinks opponent has of them

	if(know.count(opID)==0)
	{
		know.insert(make_pair(opID,make_pair(make_pair(dmLu,duLm),0)));
		//know.at(opID).first.first=dmLu;
		//know.at(opID).first.second=duLm;

		switch (dmLu)
		{
			case 'E':
				likingStats[1]++;
				break;
			case 'F':
				likingStats[2]++;
				break;
			case 'N':
				likingStats[0]++;
				break;
		}

		switch (duLm)
		{
			case 'E':
				likingStats[4]++;
				break;
			case 'F':
				likingStats[5]++;
				break;
			case 'N':
				likingStats[3]++;
				break;
		}
	}

	//Determine "meLIKESyou" based on past knowledge
	if (know.at(opID).first.first=='E') //I don't like you
			mLu=1;
	else
	{
		if (know.at(opID).first.first=='F') //I like you
			mLu=2;
		else //I am neutral towards you or I don't have a liking for you yet
			mLu=0;
	}

	//Determine "youLIKESme" based on past knowledge
	if (know.at(opID).first.second=='E') //I believe you do NOT like me
		uLm=1;
	else
	{
		if (know.at(opID).first.second=='F') //I believe you DO like me
			uLm=2;
		else //I believe you are neutral towards me or I don't have a liking set yet
			uLm=0;
	}

	//Agent has seen their opponent so know must be set to reflect this;
	know.at(opID).second=maxAbsent;

	//ASK GAME WHAT THE BITSETS ARE FOR EACH PERSPECTIVE

	int a, b, c, d, e, f, g, h, i, j, k, l=-1;

	if (row==true) //when player is "row"
	{
		a=gamePtr->payBits(0);		//IF COLUMN stays ROW does BETTER switching
		b=gamePtr->payBits(1);		//IF COLUMN stays ROW does WORSE switching
		c=gamePtr->payBits(2);		//IF COLUMN switches ROW does BETTER staying
		d=gamePtr->payBits(3);		//IF COLUMN switches ROW does WORSE staying
		e=gamePtr->payBits(4);		//IF BOTH switch ROW does BETTER
		f=gamePtr->payBits(5);		//IF BOTH switch ROW does WORSE
		g=gamePtr->payBits(6);		//IF ROW stays COLUMN does BETTER switching
		h=gamePtr->payBits(7);		//IF ROW stays COLUMN does WORSE switching
		i=gamePtr->payBits(8);		//IF ROW switches COLUMN does BETTER staying
		j=gamePtr->payBits(9);		//IF ROW switches COLUMN does WORSE staying
		k=gamePtr->payBits(10);		//IF BOTH switch COLUMN does BETTER
		l=gamePtr->payBits(11);		//IF BOTH switch COLUMN does WORSE
	}
	else  //when player is "column"
	{
		g=gamePtr->payBits(0);
		h=gamePtr->payBits(1);
		i=gamePtr->payBits(2);
		j=gamePtr->payBits(3);
		k=gamePtr->payBits(4);
		l=gamePtr->payBits(5);
		a=gamePtr->payBits(6);
		b=gamePtr->payBits(7);
		c=gamePtr->payBits(8);
		d=gamePtr->payBits(9);
		e=gamePtr->payBits(10);
		f=gamePtr->payBits(11);
	}

	int choiceIndex=-1;


	choiceIndex = l+k*2+j*4+i*8+h*16+g*32+f*64+e*128+d*256+c*512+b*1024+a*2048+mLu*4096+uLm*12288;
	choiceIndex = bit2pos(choiceIndex);

	if (choice.test(choiceIndex))
		return true;  //STAY: play ON status quo
	else
		return false; //SWITCH: play OFF status quo
}

int agent::agentID()
{
	return id;
}

int agent::agentRes()
{
	return resource;
}

void agent::setRes(int newResource)
{
	resource = newResource;
}

void agent::printAgent()
{
	//Print id
	cout << "Agent ID#" << id << endl;

	cout << "Age: " << age << endl;

	cout << "Choice Matrix" << endl;
	//Print choice matrix
	for (int i=0; i<(int)choice.size(); i++)
	{
		cout << choice[i];
	}

	cout << endl << endl << "Afterplay Matrix" << endl;
	for (int i=0; i<3; i++)
	{
		for (int j=0; j<3; j++)
		for (int k=0; k<3; k++)
		{
			cout << afterPlay[i][j][k] << "  ";
		}

		cout << endl;
	}
	cout << endl << endl;
}

	/*
	Agents update themselves by looking at the outcome they played to
	achieve and comparing it to the outcome that actually happened.
	In cases where there is an improvement over the expected payoff
	the agent may increase their perception of the other player.
	In cases where there is a decrease then the opposite may occur.

	FUTURE POSSIBILITY
	Include comparison of myPay and otherPay
	*/

void agent::updateAgent(game *const gamePtr, const int opID, int myPay, int otherPay, const bool row)
{
	resource = resource + myPay;

	int mBWS=0;
	int uBWS=0;
	int PL=0;

	if (row==true)
	{
		mBWS = gamePtr->sqPays[0]-myPay;
		uBWS = gamePtr->sqPays[1]-otherPay;
	}
	else
	{
		mBWS = gamePtr->sqPays[1]-myPay;
		uBWS = gamePtr->sqPays[0]-otherPay;
	}

	if (mBWS > 0)  //If this is true then current agent does worse than at sq
		mBWS = 0;
	else
		if (mBWS < 0) // If this is true then current agent does better than at sq
			mBWS = 2;
		else		  // If this is true then current agent is indifferent between outcome and sq
			mBWS = 1;

	if (uBWS > 0)  //If this is true then other agent does worse than at sq
		uBWS = 0;
	else
		if (uBWS < 0) // If this is true then other agent does better than at sq
			uBWS = 2;
		else		  // If this is true then other agent is indifferent between outcome and sq
			uBWS = 1;

	//Retrieve my "Previous Liking" of you
	if (know.at(opID).first.first=='E') //I didn't like you
			PL=1;
	else
	{
		if (know.at(opID).first.first=='F') //I liked you
			PL=2;
		else //I was neutral towards you
			PL=0;
	}

	//Update mLu and liking stats if necessary
	if (know.at(opID).first.first!=afterPlay[mBWS][uBWS][PL])
	{
		if (know.at(opID).first.first=='E')
			likingStats[1]--;
		else
			if (know.at(opID).first.first=='F')
				likingStats[2]--;
			else
				likingStats[0]--;

		if (afterPlay[mBWS][uBWS][PL]=='E')
			likingStats[1]++;
		else
			if (afterPlay[mBWS][uBWS][PL]=='F')
				likingStats[2]++;
			else
				likingStats[0]++;

		know.at(opID).first.first=afterPlay[mBWS][uBWS][PL];
	}


	//Retreive my perception of your "Previous Liking" of me
	if (know.at(opID).first.second=='E') //I believed you did NOT like me
		PL=1;
	else
	{
		if (know.at(opID).first.second=='F') //I believed you liked me
			PL=2;
		else //I believed you were neutral towards me
			PL=0;
	}

	//Update uLm and liking stats if necessary
	if (know.at(opID).first.second!=afterPlay[uBWS][mBWS][PL])
	{
		if (know.at(opID).first.second=='E')
			likingStats[4]--;
		else
			if (know.at(opID).first.second=='F')
				likingStats[5]--;
			else
				likingStats[3]--;

		if (afterPlay[uBWS][mBWS][PL]=='E')
			likingStats[4]++;
		else
			if (afterPlay[mBWS][uBWS][PL]=='F')
				likingStats[5]++;
			else
				likingStats[3]++;

		know.at(opID).first.second=afterPlay[uBWS][mBWS][PL];
	}

}

void agent::agentSpawn(agent* mom, agent* dad, gsl_rng* r)
{

	//correct choice pattern to reflect parental influences
	for (int i=0; i<(int)choice.size(); i++)
	{
		if (gsl_rng_uniform (r)>0.5) //take particular bitsetting from mom
		{
			if (mom->choiceBitTest(i))
				choice.set(i);
			else
				choice.reset(i);
		}
		else  //take particular bitsetting from dad
		{
			if (dad->choiceBitTest(i))
				choice.set(i);
			else
				choice.reset(i);
		}
	}

	//correct afterPlay pattern to reflect parental influences
	for (int i=0; i<3; i++)
	for (int j=0; j<3; j++)
	for (int k=0; k<3; k++)
	{
		if (gsl_rng_uniform (r)>0.5) //take particular character from mom
			afterPlay[i][j][k]= mom->afterPlayChar(i,j,k);
		else  //take particular character from dad
			afterPlay[i][j][k]= dad->afterPlayChar(i,j,k);
	}

	//correct random spawnPref pattern to reflect parental influences
	for (int i=0; i<int(spawnPref.size()); i++)
	{
		if (gsl_rng_uniform (r) <= 0.5)
		{
			if (mom->spawnPrefSet(i))
				spawnPref.set(i);
			else
				spawnPref.reset();
		}
		else
		{
			if (dad->spawnPrefSet(i))
				spawnPref.set(i);
			else
				spawnPref.reset();
		}
	}

	//correct random playPref pattern to reflect parental influences
	for (int i=0; i<int(playPref.size()); i++)
	{
		if (gsl_rng_uniform (r) <= 0.5)
		{
			if (mom->playPrefSet(i))
				playPref.set(i);
			else
				playPref.reset();
		}
		else
		{
			if (dad->playPrefSet(i))
				playPref.set(i);
			else
				playPref.reset();
		}
	}

	//correct random default likings to reflect parental influences
	if (gsl_rng_uniform (r) <= 0.5)
		dmLu=mom->getDefaultMLU();
	else
		dmLu=dad->getDefaultMLU();

	if (gsl_rng_uniform (r) <= 0.5)
		duLm=mom->getDefaultULM();
	else
		duLm=dad->getDefaultULM();

}

bool agent::choiceBitTest(int bit)
{
	if (choice.test(bit))
		return true;
	else
		return false;
}

unsigned char agent::afterPlayChar(const int i, const int j, const int k)
{
	return afterPlay[i][j][k];
}

int agent::afterPlayValues(const int j, const int k, const int l)
{
	int apv=0;

	switch (afterPlay[j][k][l])
	{
		case 'F':
			apv=1;
			break;
		case 'E':
			apv=-1;
			break;
		default:
			break;
	}
	return apv;
}

int agent::agentAge()
{
	return age;
}

void agent::agentAgePlus()
{
	age++;
}

int agent::agentBirthday()
{
	return birthday;
}

int agent::agentBitsSet()
{
	return bitsSet;
}

int agent::choiceSize()
{
	return int(choice.size());
}

bool agent::spawnAgree(const int opID)
{

	//Determine "meLIKESyou" based on past knowledge
	if (know.count(opID)==0)
		know.insert(make_pair(opID,make_pair(make_pair(dmLu,duLm),0)));

	switch (know.at(opID).first.first)
	{
		case 'E':
			switch (know.at(opID).first.second)
			{
				case 'E':
					if (spawnPref.test(0)==true) //I spawn with agents I don't like and who I think don't like me
						return true;
					break;
				case 'F':
					if (spawnPref.test(2)==true) //I spawn with agents I don't like and who I think like me
						return true;
					break;
				case 'N':
					if (spawnPref.test(1)==true) //I spawn with agents I don't like and who I think are neutral towards me
						return true;
					break;
			}
		case 'F':
			switch (know.at(opID).first.second)
			{
				case 'E':
					if (spawnPref.test(6)==true) //I spawn with agents I like and who I think don't like me
						return true;
					break;
				case 'F':
					if (spawnPref.test(8)==true) //I spawn with agents I like and who I think like me
						return true;
					break;
				case 'N':
					if (spawnPref.test(7)==true) //I spawn with agents I like and who I think are neutral towards me
						return true;
					break;
			}
		case 'N':
			switch (know.at(opID).first.second)
			{
				case 'E':
					if (spawnPref.test(3)==true) //I spawn with agents I am neutral towards and who I think don't like me
						return true;
					break;
				case 'F':
					if (spawnPref.test(5)==true) //I spawn with agents I am neutral towards and who I think like me
						return true;
					break;
				case 'N':
					if (spawnPref.test(4)==true) //I spawn with agents I like and who I think are neutral towards me
						return true;
					break;
			}
	}

	return false; //I don't spawn with agents that I like like I like you and who I think like me like I think you like me!

}

bool agent::spawnPrefSet(const int bit)
{
	if (spawnPref.test(bit)==true)
		return true;
	else
		return false;
}

bool agent::playAgree(const int opID)
{

	//Determine "meLIKESyou" based on past knowledge
	if (know.count(opID)==0)
		know.insert(make_pair(opID,make_pair(make_pair(dmLu,duLm),0)));

	switch (know.at(opID).first.first)
	{
		case 'E':
			switch (know.at(opID).first.second)
			{
				case 'E':
					if (playPref.test(0)==true) //I play with agents I don't like and who I think don't like me
						return true;
					break;
				case 'F':
					if (playPref.test(2)==true) //I play with agents I don't like and who I think like me
						return true;
					break;
				case 'N':
					if (playPref.test(1)==true) //I play with agents I don't like and who I think are neutral towards me
						return true;
					break;
			}
		case 'F':
			switch (know.at(opID).first.second)
			{
				case 'E':
					if (playPref.test(6)==true) //I play with agents I like and who I think don't like me
						return true;
					break;
				case 'F':
					if (playPref.test(8)==true) //I play with agents I like and who I think like me
						return true;
					break;
				case 'N':
					if (playPref.test(7)==true) //I play with agents I like and who I think are neutral towards me
						return true;
					break;
			}
		case 'N':
			switch (know.at(opID).first.second)
			{
				case 'E':
					if (playPref.test(3)==true) //I play with agents I am neutral towards and who I think don't like me
						return true;
					break;
				case 'F':
					if (playPref.test(5)==true) //I play with agents I am neutral towards and who I think like me
						return true;
					break;
				case 'N':
					if (playPref.test(4)==true) //I play with agents I like and who I think are neutral towards me
						return true;
					break;
			}
	}

	return false; //I don't play with agents that I like like I like you and who I think like me like I think you like me!

}

bool agent::playPrefSet(const int bit)
{
	if (playPref.test(bit)==true)
		return true;
	else
		return false;
}

unsigned char agent::getDefaultMLU()
{
	return dmLu;
}

unsigned char agent::getDefaultULM()
{
	return duLm;
}

bool agent::agentKnow(const int testAgent)
{
	if (know.count(testAgent)>=1)
		return 1;
	else
		return 0;
}

short agent::agentLike(const int knowPos, const bool fs)
{
	if (fs==true)
		if (know.at(knowPos).first.first=='E')
			return 1;
		else
			if (know.at(knowPos).first.first=='F')
				return 5;
			else
				return 3;
	else
		if (know.at(knowPos).first.second=='E')
			return 2;
		else
			if (know.at(knowPos).first.second=='F')
				return 6;
			else
				return 4;
}

int agent::agentsKnown()
{
	return (int)know.size();
}

int agent::bit2pos (const int bit)
{
	int q1=0, q2=0, q3=0, q4=0, q5=0, q6=0;

	q1=bit/4;
	q2=bit/16;
	q3=bit/64;
	q4=bit/256;
	q5=bit/1024;
	q6=bit/4096;
	return bit-q1-q2*3-q3*9-q4*27-q5*81-q6*243;
}


int agent::pos2bit (const int position)
{
	int q1=0, q2=0, q3=0, q4=0, q5=0, q6=0;

	q1=position/3;
	q2=position/9;
	q3=position/27;
	q4=position/81;
	q5=position/243;
	q6=position/729;
	return position+q1+q2*4+q3*16+q4*64+q5*256+q6*1024;
}

int agent::likingStatsCount(const int index)
{
	return likingStats[index];
}

void agent::cleanKnow(vector<int>* deadIDsPtr, const int agentCount)
{
	for(int i=0; i<agentCount; i++)
	{
		if (know.count(i)>0)
		{
			if ( (int) count(deadIDsPtr->begin(), deadIDsPtr->end(), i) > 0 || know.at(i).second==0)
				know.erase(i);
			else
				know.at(i).second--;

		}
	}
}

agent::~agent()
{

}

/*  Original work on having agents play games
	//PLAY GAMES

	//Each indexed pair of agents plays the correspondingly indexed game
	//(e.g. allMatches(3) plays allGames(3) )
	// ROW player is the first player in the pair

	int rowPlay=-1; //what the row player plays at the end
	int colPlay=-1;	//what the column player plays at the end
	int qid; 		//queue position in allAgents of agent
	int qidu;		//queue position in allAgents of opponent
	int mLu;		//Knowledge agent has of other agent
	int uLm;		//Knowledge agent thinks opponent has of them
	int bUS;		//Better for player if unilaterally switch?
	int bUSu;		//Better for opponent if unilaterally switch?
	int bMS;		//Better for both if switch off status quo
	int wUS;		//Worse for player if unilaterally switch?
	int wUSu;		//Worse for opponent if unilaterally switch?
	int wMS;		//Worse for both if mutual switch?
	int gameSQ;		//status quo cell for game being played
	int game=0;		//game in allGames agents are playing

	for(int i=0; i<allMatches.size(); i++)
	{
		for(int j=0; j<2; j++)
		{
			switch (j)
			{
				case 0:  //Determines the play for the Row player
				{
					qid = allMatches[i].players.first;
					qidu = allMatches[i].players.second;

					if (allAgents[qid].know[qidu].test(0))
						if (allAgents[qid].know[qidu].test(2))
							if (allAgents[qid].know[qidu].test(1))
								mLu=4;
							else
								mLu=3;
						else
							if (allAgents[qid].know[qidu].test(1))
								mLu=0;
							else
								mLu=1;
					else
						mLu=3;

					if (allAgents[qid].know[qidu].test(3))
						if (allAgents[qid].know[qidu].test(5))
							if (allAgents[qid].know[qidu].test(4))
								uLm=4;
							else
								uLm=3;
						else
							if (allAgents[qid].know[qidu].test(4))
								uLm=0;
							else
								uLm=1;
					else
						uLm=3;

					gameSQ=allGames[i].sq.to_ulong();

					Need to set choice values to return play
					bUS = allGames[i].pays[gameSQ] < allGames[i].pays[gameSQ];

					rowPlay = allAgents[qid].choice[mLu][uLm][;
					break;
				}
				case 1: //Determines the play for the Column player
			}
			game++;
		}
		Take i and use it to index allMatches and allGames
		new for loop to switch between first/row and second/column
		take payoff info for allGames and use it to determine a play
		based on index value to allAgents stored in allMatches
		Store the payoffs in an array
		print them out
	}

	//Update agent prefs, etc.

		//Assign a set of base values to the agent's status quo conversion array.unsigned char
	//Currently these are basically junk values and the same for everyone.
	//"Sensible" randomness needs to be built in.
	int diff=0;
	unsigned char sqc='0';

	for (int i=0; i<10; i++)
	{
		for (int j=0; j<10; j++)
		{
			diff=i-j;
			sqc=6;

			switch (diff)
			{
				case -9:
					sqc='5';
					break;
				case -8:
					sqc='5';
					break;
				case -7:
					sqc='5';
					break;
				case -6:
					sqc='5';
					break;
				case -5:
					sqc='5';
					break;
				case -4:
					sqc='4';
					break;
				case -3:
					sqc='4';
					break;
				case -2:
					sqc='4';
					break;
				case -1:
					sqc='4';
					break;
				case 0:
					sqc='3';
					break;
				case 1:
					sqc='3';
					break;
				case 2:
					sqc='3';
					break;
				case 3:
					sqc='2';
					break;
				case 4:
					sqc='2';
					break;
				case 5:
					sqc='2';
					break;
				case 6:
					sqc='1';
					break;
				case 7:
					sqc='1';
					break;
				case 8:
					sqc='1';
					break;
				case 9:
					sqc='1';
					break;
			}
			sqCode[i][j]= sqc;
		}
	}

*/
