/* This work is licensed under the Creative Commons Attribution-Share Alike 2.5 Canada License.
 * This To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/2.5/ca/
 * This or send a letter to Creative Commons, 171 Second Street, Suite 300, San Francisco,
 * This California, 94105, USA.
 *
 * Written by John Simpson, Department of Philosophy, University of Alberta, 2009, 2010.
*/

#include <iostream>
#include "game.h"

using namespace std;

/*
I also need (maybe?) a 'switch' somewhere to turn-on/off symmetry b/w agent payoffs
Currently symmetry is possible, but not forced

All payoffs are 0-9
*/

		/* This is the standard 2x2 game matrix with 4 outcomes and 8 payoffs.
		Convention:
		pays[0]=payoff to player A for outcome T,L
		pays[1]=payoff to player A for outcome B,L
		pays[2]=payoff to player A for outcome T,R
		pays[3]=payoff to player A for outcome B,R
		pays[4]=payoff to player B for outcome T,L
		pays[5]=payoff to player B for outcome T,R
		pays[6]=payoff to player B for outcome B,L
		pays[7]=payoff to player B for outcome B,R
		*/

game::game(const int type, gsl_rng * r)
{

	gameType = type;

	switch (gameType)
		{
			case 0:
				RNDgame(r);
				break;
			case 1:
				PDgame(r);
				break;
			case 2:
				HNTgame(r);
				break;
			case 3:
				CHKgame(r);
				break;
			case 4:
				BSgame(r);
				break;
		}

	//Collect paySum information
	paySum=0;
	for (int i=0; i<8; i++)
		paySum=paySum+pays[i];

	//sq.set(0); //TEST to see if setting sq always to T-L makes a difference
	//sq.set(1);


	//Assign a random status quo
	if (gsl_rng_uniform (r) > 0.5)
		sq.flip(0); //Sets TOP when set
	if (gsl_rng_uniform (r) > 0.5)
		sq.flip(1); //Sets LEFT when set


	//Determine better/worse unilateral/mutual switch values
	//At the same time, collect status quo payoffs
	//This information is used by agents to determine what kind of
	//game they are playing when agent::play is called

	if (sq.test(0))  //TOP
		if (sq.test(1))  //TOP-LEFT
			{
			sqPays[0]=pays[0];
			sqPays[1]=pays[4];

			//Does row do better or worse if they switch and column stays?
			if (pays[0]<pays[1])  //better
				bw.set(0);
			else
				if (pays[0]>pays[1])  //worse
					bw.set(1);

			//Does row do better or worse if column unilaterally switches?
			if (pays[0]<pays[2])  //better
				bw.set(2);
			else
				if (pays[0]>pays[2])  //worse
					bw.set(3);

			//Does row do better or worse if they both switch?
			if (pays[0]<pays[3])  //better
				bw.set(4);
			else
				if (pays[0]>pays[3])  //worse
					bw.set(5);

			//Does column do better or worse if they switch and row stays?
			if (pays[4]<pays[5])  //better
				bw.set(6);
			else
				if (pays[4]>pays[5])  //worse
					bw.set(7);

			//Does column do better or worse if row switches and they stay?
			if (pays[4]<pays[6])  //better
				bw.set(8);
			else
				if (pays[4]>pays[6])  //worse
					bw.set(9);

			//Does column do better or worse if they both switch?
			if (pays[4]<pays[7])  //better
				bw.set(10);
			else
				if (pays[4]>pays[7])  //worse
					bw.set(11);
			}

			else  //TOP-RIGHT
			{
			sqPays[0]=pays[2];
			sqPays[1]=pays[5];

			//Does row do better or worse if they switch and column stays?
			if (pays[2]<pays[3])  //better
				bw.set(0);
			else
				if (pays[2]>pays[3])  //worse
					bw.set(1);

			//Does row do better or worse if column unilaterally switches?
			if (pays[2]<pays[0])  //better
				bw.set(2);
			else
				if (pays[2]>pays[0])  //worse
					bw.set(3);

			//Does row do better or worse if they both switch?
			if (pays[2]<pays[1])  //better
				bw.set(4);
			else
				if (pays[2]>pays[1])  //worse
					bw.set(5);

			//Does column do better or worse if they switch and row stays?
			if (pays[5]<pays[4])  //better
				bw.set(6);
			else
				if (pays[5]>pays[4])  //worse
					bw.set(7);

			//Does column do better or worse if row switches and they stay?
			if (pays[5]<pays[7])  //better
				bw.set(8);
			else
				if (pays[5]>pays[7])  //worse
					bw.set(9);

			//Does column do better or worse if they both switch?
			if (pays[5]<pays[6])  //better
				bw.set(10);
			else
				if (pays[5]>pays[6])  //worse
					bw.set(11);
			}
	else  //BOTTOM
		if (sq.test(1))  //BOTTOM-LEFT
			{
			sqPays[0]=pays[1];
			sqPays[1]=pays[6];

			//Does row do better or worse if they switch and column stays?
			if (pays[1]<pays[0])  //better
				bw.set(0);
			else
				if (pays[1]>pays[0])  //worse
					bw.set(1);

			//Does row do better or worse if column unilaterally switches?
			if (pays[1]<pays[3])  //better
				bw.set(2);
			else
				if (pays[1]>pays[3])  //worse
					bw.set(3);

			//Does row do better or worse if they both switch?
			if (pays[1]<pays[2])  //better
				bw.set(4);
			else
				if (pays[1]>pays[2])  //worse
					bw.set(5);

			//Does column do better or worse if they switch and row stays?
			if (pays[6]<pays[7])  //better
				bw.set(6);
			else
				if (pays[6]>pays[7])  //worse
					bw.set(7);

			//Does column do better or worse if row switches and they stay?
			if (pays[6]<pays[4])  //better
				bw.set(8);
			else
				if (pays[6]>pays[4])  //worse
					bw.set(9);

			//Does column do better or worse if they both switch?
			if (pays[6]<pays[5])  //better
				bw.set(10);
			else
				if (pays[6]>pays[5])  //worse
					bw.set(11);
			}
		else  //BOTTOM-RIGHT
			{
			sqPays[0]=pays[3];
			sqPays[1]=pays[7];

			//Does row do better or worse if they switch and column stays?
			if (pays[3]<pays[2])  //better
				bw.set(0);
			else
				if (pays[3]>pays[2])  //worse
					bw.set(1);

			//Does row do better or worse if column unilaterally switches?
			if (pays[3]<pays[1])  //better
				bw.set(2);
			else
				if (pays[3]>pays[1])  //worse
					bw.set(3);

			//Does row do better or worse if they both switch?
			if (pays[3]<pays[0])  //better
				bw.set(4);
			else
				if (pays[3]>pays[0])  //worse
					bw.set(5);

			//Does column do better or worse if they switch and row stays?
			if (pays[7]<pays[6])  //better
				bw.set(6);
			else
				if (pays[7]>pays[6])  //worse
					bw.set(7);

			//Does column do better or worse if row switches and they stay?
			if (pays[7]<pays[5])  //better
				bw.set(8);
			else
				if (pays[7]>pays[5])  //worse
					bw.set(9);

			//Does column do better or worse if they both switch?
			if (pays[7]<pays[4])  //better
				bw.set(10);
			else
				if (pays[7]>pays[4])  //worse
					bw.set(11);
			}

}

void game::printGame()
{
	cout << "Game type:" << gameType << endl;

	cout << pays[0] << "   " << pays[4] << "   " << "|" << "   " << pays[2] << "   " << pays[5] << "   " << endl;
	cout << "----------------" << endl;
	cout << pays[1] << "   " << pays[6] << "   " << "|" << "   " << pays[3] << "   " << pays[7] << "   " << endl;

	cout << endl;
	cout << "With Status Quo Outcome:" << endl;
	if (sq.test(0))  //TOP
		if (sq.test(1))  //TOP-LEFT
			cout << "TOP-LEFT";
		else
			cout << "TOP-RIGHT";
	else  //BOTTOM
		if (sq.test(1))  //BOTTOM-LEFT
			cout << "BOTTOM-LEFT";
		else  //BOTTOM-RIGHT
			cout << "BOTTOM-RIGHT";

	cout << endl << endl;
}

int game::payBits(const int bitTest)
{
	int bit=0;
	if (bw.test(bitTest))
		bit=1;
	return bit;
}

int game::getPaySum()
{
	return paySum;
}

int game::outcome(const bool row, const bool rowPlay, const bool colPlay)
{
	if (row==true)  //Return row player's payoff on the outcome
		if (sq.test(0))  //TOP
			if (sq.test(1))  //TOP-LEFT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[0];
					else
						return pays[2];
				else
					if (colPlay==true)
						return pays[1];
					else
						return pays[3];
			else  //TOP-RIGHT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[2];
					else
						return pays[0];
				else
					if (colPlay==true)
						return pays[3];
					else
						return pays[1];
		else  //BOTTOM
			if (sq.test(1))  //BOTTOM-LEFT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[1];
					else
						return pays[3];
				else
					if (colPlay==true)
						return pays[0];
					else
						return pays[2];
			else  //BOTTOM-RIGHT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[3];
					else
						return pays[1];
				else
					if (colPlay==true)
						return pays[2];
					else
						return pays[0];
	else  //Return column player's payoff on the outcome
		if (sq.test(0))  //TOP
			if (sq.test(1))  //TOP-LEFT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[4];
					else
						return pays[5];
				else
					if (colPlay==true)
						return pays[6];
					else
						return pays[7];
			else  //TOP-RIGHT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[5];
					else
						return pays[4];
				else
					if (colPlay==true)
						return pays[7];
					else
						return pays[6];
		else  //BOTTOM
			if (sq.test(1))  //BOTTOM-LEFT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[6];
					else
						return pays[7];
				else
					if (colPlay==true)
						return pays[4];
					else
						return pays[5];
			else  //BOTTOM-RIGHT
				if (rowPlay==true)
					if (colPlay==true)
						return pays[7];
					else
						return pays[6];
				else
					if (colPlay==true)
						return pays[5];
					else
						return pays[4];
}

void game::RNDgame(gsl_rng * r)
{
	//Build game with random payoffs 0 - 9
	for (int i=0; i<8; i++)
		pays[i]=(int)(gsl_rng_uniform_int(r,10));
}

//This version builds PDs
void game::PDgame(gsl_rng * r)
{
	//Sets "cooperative" payoffs, used as ref.pt. for others: 3 to 8
	pays[0]=(int)(3+gsl_rng_uniform_int(r,6));
	pays[4]=(int)(3+gsl_rng_uniform_int(r,6));

	//Sets "unilateral defection" payoffs: coop pay to 9
	pays[1]=(int)((pays[0]+1)+gsl_rng_uniform_int(r,(9-pays[0]+1)));
	pays[5]=(int)((pays[4]+1)+gsl_rng_uniform_int(r,(9-pays[4]+1)));

	//Sets "mutual defection" payoff: 2 to (coop pay - 1)
	pays[3]=(int)(1+gsl_rng_uniform_int(r,(pays[0]-2)));
	pays[7]=(int)(1+gsl_rng_uniform_int(r,(pays[4]-2)));

	//Sets "unilateral defection against" payoffs:
	pays[2]=(int)(gsl_rng_uniform_int(r,(pays[3])));
	pays[6]=(int)(gsl_rng_uniform_int(r,(pays[7])));

}

void game::HNTgame(gsl_rng * r) //This version builds Stag Hunts (with Tragedy of commons loss)
{

	//Set "Stag together" payoffs, used as ref.pt for others: 3 to 9
	pays[0]=(int)(3+gsl_rng_uniform_int(r,7));
	pays[4]=(int)(3+gsl_rng_uniform_int(r,7));

	//Set "Unilateral Rabbit Hunting" payoffs: 2 to (Stag together - 1)
	pays[1]=(int)(2+gsl_rng_uniform_int(r,pays[0]-2));
	pays[5]=(int)(2+gsl_rng_uniform_int(r,pays[4]-2));

	//Set "Stag alone" payoffs: 0 to (uRabbit - 2)
	pays[2]=(int)(gsl_rng_uniform_int(r,pays[1]-1));
	pays[6]=(int)(gsl_rng_uniform_int(r,pays[5]-1));

	//Set "WABBITS!" payoffs: (sAlone + 1) to (uRabbits - 1)
	pays[3]=(int)(pays[2]+1+gsl_rng_uniform_int(r,(pays[1]-pays[2]-2)));
	pays[7]=(int)(pays[6]+1+gsl_rng_uniform_int(r,(pays[5]-pays[6]-2)));
}

void game::CHKgame(gsl_rng * r)
{

	int stage=0;

	//This version builds Chickens
	for (int i=0; i<2; i++)
		for (int j=0; j<4; j++)
		{
			stage=i*4;
			switch (j)
			{
				case 0: //Sets "cooperative"  payoffs, used as ref.pt for others: 2 to 8
					pays[stage+j]=(int)(2+gsl_rng_uniform_int(r,9));
					break;
				case 1: //Sets "unilaterally straight" payoffs: coop pay to 9
					pays[stage+j]=(int)((pays[stage+j-1]+1)+(gsl_rng_uniform_int(r,(9-(pays[stage+j-1]+1)))));
					break;
		/* NEEDS FIXING
				case 2: //Sets "unilaterally swerve" payoffs: 1 to (coop pay - 1)
					pays[stage+j]=(int)(1+((pays[stage+j-2]-1)*rnd->getRand())+0.5);
					break;
				case 3: //Sets "mutually straight" payoffs: 0 to u-swerve pay
					pays[stage+j]=(int)(((pays[stage+j-1])*rnd->getRand())+0.5);
					break;
		*/
			}
		}
}

//This version builds BoSs
void game::BSgame(gsl_rng * r)
{
	/* NEEDS FIXING
	singleRand* rnd = singleRand::getInstance();

	//Sets "together at ideal"  payoffs, used as ref.pt for others: 3 to 9
	pays[0]=(int)(3+((6)*rnd->getRand())+0.5);
	pays[7]=(int)(3+((6)*rnd->getRand())+0.5);

	//Sets "together at other" payoffs: 2 to (ideal - 1)
	pays[3]=(int)(2+((pays[0]-3)*rnd->getRand())+0.5);
	pays[4]=(int)(2+((pays[7]-3)*rnd->getRand())+0.5);

	//Sets "alone at favourite" payoffs: 1 to (other - 1)
	pays[2]=(int)(1+((pays[3]-2)*rnd->getRand())+0.5);
	pays[5]=(int)(1+((pays[4]-2)*rnd->getRand())+0.5);

	//Sets "alone at other" payoffs: 0 to (alone at fav - 1)
	pays[1]=(int)(0+((pays[2]-1)*rnd->getRand())+0.5);
	pays[6]=(int)(0+((pays[5]-1)*rnd->getRand())+0.5);
	*/
}

game::~game()
{
}
